# -*- coding: utf-8 -*-

# utility for updating effectdata in Telescope v2
# author: cag
# license: CC BY-NC-SA 4.0

import os, sys, io, shutil, re
import zipfile, tempfile, datetime
import getpass
import xml.parsers.expat
# import plistlib, xml.parsers.expat
from operator import attrgetter, itemgetter

Python2 = sys.version_info[0] == 2
if Python2:
	import ConfigParser as configparser
	from plistlib import readPlist as readPlistFile
	from plistlib import readPlistFromString as readPlistStream
else:
	import configparser
	from plistlib import load as readPlistFile
	from plistlib import loads as readPlistStream
	
from platform import system as platform_system
curr_system = platform_system()
platformIsLinux = curr_system == 'Linux'
platformIsWindows = curr_system == 'Windows'
platformIsMacOS = curr_system not in ['Linux', 'Windows']

#####################################
## toggle all before mk

# DEBUG = False
DEBUG = True

import pdb
import traceback

def fmtVal(val, depth=0):
	try:
		if val is None:
			return val

		if isinstance(val, (Entity, Property)):
			return fmtVal(val.__dict__, depth+1)

		indent = '  '*depth
		newLine = '\n{}'.format(indent)
		if isinstance(val, dict):
			comma = ',{}'.format(newLine)
			fmt = newLine + '{'
			for x, y in val.items():
				fmt += ' {}: {}'.format(x, fmtVal(y, depth+1))
				fmt += comma
				fmt += ' '
			return fmt[:-1] + '}'

		if isinstance(val, list):
			comma = ', '
			if any(isinstance(x, (list, dict)) for x in val):
				comma = ',{}'.format(newLine)
			return '[{}]'.format(
					comma.join(fmtVal(x, depth+1)
						for x in val))

		if isinstance(val, str):
			if '\n' in val:
				return '"{!r}"'.format(val)
			return '"{}"'.format(val)

		return str(val)
	except Exception as exc:
		print('fmtVal, {}'.format(exc))
		traceback.print_exc()
		pdb.set_trace()

def prtVal(val):
	print(fmtVal(val))

#####################################

# OPTIONS
								
RemoveFlasherSubEnts = True     # to reduce output size (& since telescope collects light over time...  \_0_/ )
LogFoldersSearched = True       # log all folders that are scanned by walker
logFoldersProcessed = True		# log folder where INPUT_FN where found and loaded
LogOxzSearched = True			# log all folders that are scanned by walker
logOxzProcessed = True			# log folder where INPUT_FN where found and loaded
AuditUniformTextures = True     # compare uniform indices to textures list to find missing textures
LogMissingSemiColon = False      # reporting of missing ';'s after lists & dictionaries
LogUnquotedDash = False          # reporting of unquoted keys/values containing '-'s (there are lots!)
LogDuplicateKeys = False         # reporting of keys duplicated across different oxp's
LogOnlyDifferentDups = True     # limit logging of duplicates to those that differ vs. logging all
								# are no longer output, so must add to 'Keys_to_output' in addition to setting this True

# keys to output in .plist; an empty [] will output all
Keys_to_output = [ # keys for effectdata.plist files; removed 'script', 'script_info' as not interchangable
	'beacon', 
	'is_break_pattern', 
	'is_external_dependency', # "is_external_dependency" = yes; suppress errors; ?add if not present
	'is_template', 
	'like_effect', 
	'like_ship', # 'like_ship' added to get swapped for 'like_effect'; both required as script also used to process existing effects data
	'materials', 
	'model', 
	'scanner_display_color1', 
	'scanner_display_color2',
	# 'shaders', safer to rely on only materials; would need to exclude some uniforms as some are dynamic
	#  eg. taxi_adder from taxi_galactica_2.1.oxz cycles through various PaintColor's when first aquired!
	'smooth', 
	'subentities' 
]

# keys are renamed before output  
Keys_to_rename = { 'like_ship': 'like_effect' } # convert shipdata to effectdata requires renaming keys; ?only one

# keys to insert when a (trigger) key is detected
# dict of dict w/ each entry being triggerKey: { insertKey: insertValue }
Keys_to_insert = { 'like_effect': { 'is_external_dependency': 'yes' },
	# add 'is_external_dependency' to suppress oolite errors for ref's to files in missing oxp's
}

# keys that are not suppressed due to a like_ship property
Keys_not_suppressed = ['is_template']
Keys_not_suppressed.extend(Keys_to_rename.values())
# Keys_not_suppressed.extend(list(inserts.keys() for inserts in Keys_to_insert.values()))
Keys_not_suppressed.extend(inKey for inserts in Keys_to_insert.values()
					  		for inKey, inVal in inserts.items())
# subkeys to insert when a (trigger) key is detected and key is not already present
# dict of dict w/ each entry being triggerKey: { insertKey: insertValue,... }
Subkeys_to_insert_if_empty = { 
	'materials': { 'illumination_map': 'telescope-illumination.png',
				   'illumination_modulate_color': 'blackColor' }
}

# these keys are output on a single line
# subentities are a special case, as there is a legacy format of simple strings that are not compacted
Keys_to_compact = [ 'scanner_display_color1', 'scanner_display_color2', 
					'materials', 'shaders', 'subentities' ]

# these subkeys are output on a single line
Subkeys_to_compact = [ 'textures', 'uniforms' ]

OXP_FOLDER_EXT = '.oxp'	 # ext of folders searched (must be LOWERCASE!)
OXZ_FILE_EXT = '.oxz'	 # ext of files searched (must be LOWERCASE!)
ZIPPEDFILES = [ '.oxz' ] # ext of zip files searched (must be LOWERCASE!)
# - to incl .zip files use: [ '.oxz', '.zip' ]
# if [], all zip type files, regardless of .ext, will be checked

INPUT_FILE = 'shipdata'
INPUT_EXT = '.plist'
INPUT_FN = INPUT_FILE + INPUT_EXT
INPUT_FOLDER = 'Config'

OUTPUT_FILE = 'effectdata'
OUTPUT_EXT = '.plist'
OUTPUT_FN = OUTPUT_FILE + OUTPUT_EXT
OUTPUT_FOLDER = 'Config'

LOG_FILE = 'collect_shipdata'
LOG_EXT = '.log'
LOG_FN = LOG_FILE + LOG_EXT
LOG_MESSAGES = True
LOG_WARNINGS = True
LOG_ERRORS = True
LOG_LINELEN = 100
MAX_FOLDER_DEPTH = 20

MANIFEST_FN = 'manifest.plist'
TELESCOPE_ID = 'oolite.oxp.Norby.cag.Telescope'
REQUIRED_FOLDER = 'effect data'
REQUIRED_FN = 'effectdata.required.plist'
IDENTIFIER_RE = re.compile(r'''^\s*(?P<kQ>['"]?)identifier(?P=kQ)\s*=\s*(?P<vQ>['"])(?P<id>.*?)(?P=vQ)''')

NL = '\n'
TAB = '\t'
TABSPACE = 2
OPENINGS = '([{'
CLOSINGS = ')]}'
ENCLOSURES = OPENINGS + CLOSINGS
QUOTES = '\'"`'   
OUTPUT_FILE_HEADER = '{' + NL
OUTPUT_FILE_TAIL = '}' + NL
TRUEVALUES = [ 'true', 'yes', 'on' ]
FALSEVALUES = [ 'false', 'no', 'off' ]
BOOLEANS = TRUEVALUES + FALSEVALUES

TRANS_CHARS = '"\';()[]{}\n\f\r\t\v'
# used in str.translate calls in _str_minimize()
if Python2:	#  'alphabet code'.translate(None, 'abc')
	from string import maketrans
	MakeSparse = maketrans( TRANS_CHARS, ' ' * len( TRANS_CHARS ) ) 
else:		#  'alphabet code'.translate(''.maketrans({'a': '', 'b':'', 'c':''}))
	MakeSparse = ''.maketrans( TRANS_CHARS, ' ' * len( TRANS_CHARS ) ) 
	
ORIGIN_MARKER = 'sDpL:' # part of re, so careful what you set it to!
DUP_KEY_PREFIX = '__dupKey_'

# GLOBAL VARIABLES

Launch_dir = None       # full path of current working directory
Input_file_count = 0    # # of files processed by tree crawler
Input_size = 0          # size of all input INPUT_FILE's
Required_size = 0		# size of telescope's essential data
Output_size = 0         # size of resultant data
CurrentFile = None      # filename for error msgs
Logfile = None
TelescopeEffectData = '' # required telescope effectdata (ie. native to that oxp)
CurrentEffectData = '' 	# existing telescope effectdata
Files_logged = []       # files which have had an error or warning (so only output once)
Registered_ents = dict()
SourceFileNames= []
TelescopeOxp = None		# path to telescope oxp folder
TelescopeOxz = None		# path to telescope oxz file

## class definitions ##########################################################


class MatchFailError( Exception ):			# for Plist decoder
	"""Exception raised for failed regex match."""

	def __init__( self, index, nextchar ):
		self.index = index
		self.nextchar = nextchar
	
class Entity:
	def __init__( self, name ):
		self.name = _unquote( name )        # its name, stripped of quotes (if any) from source; always quoted on output
		self.place = []                     # array of indices in source files where entity definition starts
		self.origin = []                    # parallel list of SourceFileNames indices where it was found
		self.properties = {}                # dict of list(Property), mapping keys to list of value(s), eg. model: [] of Property
		self.output_oi = 0                  # index of origin file used to generate output
	def add_origin( self, file, place ):
		if file not in self.origin: 
			self.origin.append( file )
			self.place.append( 0 if place is None
						else int( place ) ) # index in source where entity definition starts
		return self.origin.index( file )
	def add_item( self, key, value ):
		if key in self.properties:
			self.properties[ key ].append( value )
		else:
			self.properties[ key ] = [ value ]

class Property:
	def __init__( self, key, place = 0, value = None, vplace = 0, origID = 0, inserted = False ):
		self.name = _unquote( key )         # its name, stripped of quotes, for comparison
		self.place = 0 if place is None \
					   else int( place )    # index in source where property name starts
		self.value = value                  # a Python object parsed from file sources
		self.vplace = 0 if vplace is None \
						else int( vplace )  # index in source where property definition starts
		self.plist = []                     # list of formatting strings representing its obj
		self.length = 0                     # total length of items in plist
		self.origID = 0 if origID is None \
						else int( origID )  # index of origin file
		self.inserted = inserted            # flag for those generated using Keys_to_insert
### .inserted no longer needed (got rid of <stub>)
						
## errors & IO ################################################################

def _ent_file_msg( origin, ent, dups = None ):
	try:

		def format_dups(duplist, string):
			dup_names = []
			first = True
			for dup in duplist:
				if dup not in dup_names:
					dup_names.append(dup)
				else:
					continue
				if not first:
					string += ', '
				if len(string) - string.rfind(NL) >= LOG_LINELEN:
					string += NL + '    '
				first = False
				string += dup
			return string

		reg_ent = Registered_ents[ent]
		outI = reg_ent.output_oi
		if origin not in Files_logged: # only write it once
			Files_logged.append( origin )
			msg = 'File {}:'.format(SourceFileNames[origin[outI]])
			Logfile.write(NL + msg + NL)
		files = len( origin )
		for index in range(files):
			if index == outI: continue
			msg = '  - data also found in: {}'.format(
					SourceFileNames[origin[index]])
			Logfile.write(msg + NL)
		total = 0 if dups is None else len(dups)
		if total > 0:
			output = '  Warning!  {} contain{} {} with {}duplicate definition{} for:'\
					.format('this file' if files == 2 else 'these files',
							's' if files > 2 else '',
							'an entity' if total == 1 else 'entities',
							'a ' if files == 1 else '',
							'' if files == 1 else 's' )
			output += NL + '    '
			dup_ents = list(dups.keys())
			dup_ents.sort()
			output = format_dups(dup_ents, output)
			dup_keys = []
			for dupe in dup_ents:
				# dups is a dict of ... ent: [(dup_key, orig_key)]
				prnames = dups[ dupe ]
				for prop, origp in prnames:
					dup_keys.append(origp)
			length = len(dup_keys)
			dup_keys.sort()
			total += length
			output += NL + '  key{} duplicated:'.format(
						'' if length == 1 else 's')
			output += NL + '    '
			output = format_dups(dup_keys, output)
			Logfile.write(output + NL)
	except Exception as exc:
		if DEBUG:
			print('_ent_file_msg, Error: {}'.format(exc))
			traceback.print_exc()
			pdb.set_trace()
		else:
			print('{}\n_ent_file_msg,  dups: {}'.format(exc, dups))

def _ent_dup_error( dups ):
	try:
		for ent, dup_keys in dups.items():
			reg_ent = Registered_ents[ent]
			values = reg_ent.properties # dict of key, [value]
			curr_oi = reg_ent.output_oi
			dup_keys.sort()
			reported = dict()
			for prop, origp in dup_keys:
				curr = _str_minimize(values[prop][curr_oi].value, translate=False)
				if prop not in reported.keys():
					reported[prop] = []
				prop_dups = [values[prop][i] for i in range(len(values[prop]))
							 					if i != curr_oi]
				for new in prop_dups:
					if new.value is None:  ## a stub placed in _select_origin
						continue
					dup = _str_minimize(new.value, translate=False)
					# declutter the log; already noted under keys duplicated
					if _equal_values(curr, dup):
						continue
					if dup in reported[prop]:
						continue
					else:
						reported[prop].append(dup)
					intro = '  {}[ {} ]: '.format(ent, origp)
					leni = len(intro)
					first = leni + len(curr)
					total = first + len(dup)
					spcr1 = '' if first < LOG_LINELEN else NL + '        '
					spcr2 = '' if total < LOG_LINELEN \
							else (NL + '    ') if first > LOG_LINELEN \
									else (NL + ' '*(leni - 4))
					Logfile.write(intro + '{}"{}" {} -> "{}"'.format(
											spcr1, dup, spcr2, curr) + NL)
	except Exception as exc:
		if DEBUG:
			print('_ent_dup_error, Error: {}'.format(exc))
			traceback.print_exc()
			pdb.set_trace()
		else:
			print('{}\n_ent_dup_error,  dups: {}'.format(exc, dups))

def _log_blankline():
	Logfile.write(NL)
	
def _log_message( msg ):
	if not LOG_MESSAGES: return
	_write_log( '', msg )

def _log_warning( msg ):
	if not LOG_WARNINGS: return
	_write_log( 'Warning:', msg, showfn = True )

def _log_error( msg ):
	if not LOG_ERRORS: return
	_write_log( 'Error!', msg, suffix = NL, showfn = True )

Last_logged_fn = None
Errors_in_log = 0
Warnings_in_log = 0
def _write_log( kind, msg, suffix = '', showfn = False ):
	global Last_logged_fn, Errors_in_log, Warnings_in_log

	if kind == 'Error!':
		Errors_in_log += 1
		_log_blankline()
	if kind == 'Warning:':
		Warnings_in_log += 1
		_log_blankline()
	if showfn and CurrentFile is not None:
		if CurrentFile != Last_logged_fn:
			Logfile.write('file: {}'.format(CurrentFile) + NL)
			Logfile.write('      {}'.format( '=' * len(CurrentFile)) + NL)
			Last_logged_fn = CurrentFile
	Logfile.write('{}{}{} {}{}'.format(
				kind, '' if kind == '' else ' ', msg, NL, suffix))

def _load_zipped_effects( root, file ):
	global TelescopeOxz

	try:
		err = False
		path = os.path.join( root, file )
		if zipfile.is_zipfile( path ):
			_, oxzName = os.path.split( path )
			if LogOxzSearched:
				_log_message( '  checking zip {}'.format( oxzName ) )
			with zipfile.ZipFile( path ) as zipf:
				zips = zipf.namelist()
				for zname in zips:
					zpath, zfn = os.path.split( zname )
					if len(zpath) > 0 and zpath != INPUT_FOLDER:
						# ignore folders not root and not 'Config'
						continue
					if zfn.lower() == MANIFEST_FN:
						with zipf.open(zname, mode='r') as fp:
							lines = fp.readlines()
						if getIdentifier(lines, inOxz=True) == TELESCOPE_ID:
							TelescopeOxz = path
							data = '{}{}/{}'.format(
									(zpath + '/') if len(zpath) else '',
									REQUIRED_FOLDER, REQUIRED_FN)
							with zipf.open(data, mode='r') as zp:
								loadEssentialData(zp, inOxz=True)
							data = '{}{}/{}'.format(
									(zpath + '/') if len(zpath) else '',
									OUTPUT_FOLDER, OUTPUT_FN)
							with zipf.open(data, mode='r') as zp:
								loadCurrentData(zp, inOxz=True)

					if zfn.lower() != INPUT_FN:
						continue
					with zipf.open( zname, mode = 'r' ) as fp:
						zfnStr = os.path.join(zpath, INPUT_FN)
						if logOxzProcessed and LogOxzSearched:
							_log_message('  -> loading: {}'.format(zfnStr))
						elif logOxzProcessed:
							_log_message('  -> {}: loading: {}'.format(
									oxzName, zfnStr))
						err = _load_effects( fp, path, inOxz=True ) or err
		else:
			raise ValueError( '_load_zipped_effects, not a zipfile', path )
	except Exception as exc:
		if DEBUG:
			print('_load_zipped_effects, Error: {}'.format(exc))
			traceback.print_exc()
			pdb.set_trace()
		else:
			print('{}\n_load_zipped_effects, root = "{}", file = "{}"'
				  .format(exc, root, file))
	return err
	
	
## proc interface #############################################################

	
def _load_effects( fp, file, inOxz=False ):
	global CurrentFile, Input_size, Input_file_count
	
	err = False
	plist = ''
	try:
		CurrentFile = file
		plist = fp.read()
		length = len( plist )
		if length == 0: 
			_log_error('_load_effects, failed to read file; length of 0' )
			return True
		xml_test = b'<?xml' if inOxz else '<?xml'
		if plist.startswith( xml_test ):
			Input_size += length
			if inOxz: # zipfiles return bytes, not str
				parsed = readPlistStream( plist )
			else:
				# fp.seek( 0 )
				# parsed = plistlib.load( fp, fmt = plistlib.FMT_XML  )
				fp.close()
				fp = open( file, 'rb' )
				parsed = load( fp )
		else:
			Input_size += length
			if inOxz: # zipfiles return bytes, not str
				plist = plist.decode( encoding = "utf-8", errors = "strict" )
			parsed, _ = PListDecode( plist ) # returns tuple (obj,index of next char, ie. eof)
		if parsed is not None:
			if not inOxz:
				if logFoldersProcessed and LogFoldersSearched:
					_log_message('  -> loading: {}'.format(file))
				elif logFoldersProcessed:
					folder, fn = os.path.split(file)
					_log_message('{}: loading: {}'.format(folder, fn))
			Input_file_count += 1
			if isinstance( parsed, dict ):
				_proc_entities( file, parsed )
			else:
				err = True
				_log_error('_load_effects, "{}", parsed value not a dict: {}{}'
						   .format(file, NL, parsed))
		else:
			err = True
			_log_error('_load_effects, file failed to parse!')
	except xml.parsers.expat.ExpatError as xerr:
		err = True
		if inOxz:
			plist = plist.decode( encoding = "utf-8", errors = "strict" )
		lines = plist.splitlines()
		context = errcontext( lines[ xerr.lineno - 1 ], xerr.offset )
		_log_error( "line {}, column {}: {}{}".format( 
				xerr.lineno, xerr.offset, xml.parsers.expat.errors.messages[ xerr.code ], context ) )
	except ValueError as verr:
		err = True
		_log_error('_load_effects, Error: {}'.format(verr))
	except Exception as exc:
		err = True
		if DEBUG:
			print('_load_effects, Error: {}'.format(exc))
			traceback.print_exc()
			pdb.set_trace()
		else:
			print( '{}\n_load_effects, file = "{}"'.format( exc, file ))
	finally:
		CurrentFile = None
	return err
 
# '#' + str( index ) + '#' + key + '##'  for dict keys & list values
PLACEMENT_RE = re.compile( r'''
			\# (\d+) \#                           # placement number
			   ( .*? )
			\#\#
			''', re.VERBOSE | re.DOTALL | re.MULTILINE )

def _encode_placement( info, place ): # encode 'info' with 'place'
	# placement is needed at all levels for sorting
	# this makes Registered_ents objects have a mix of encoded and decoded keys
	# (but consistent on each level)
	# - remove placement to find dups of entity keys, but not sub-keys
	return '#' + str(place) + '#' + str(info) + '##'

def _decode_placement( string ): # break string into data and placement index, del any \r
	info = string
	match = place = None
	try:
		if isinstance(string, str):
			match = PLACEMENT_RE.match(string)
			info = match.group(2) if match else info.replace('\r', '')
		elif isinstance(string, dict):
			match = min([PLACEMENT_RE.match(key) for key in string.keys()],
						key=lambda x: int(x.group(1)) if x else 1e10)
		elif isinstance(string, list):
			first = string
			while isinstance(first, list):
				if len(first) == 0:
					return info, place
				first = first[0]
			if isinstance(first, (str, dict)):
				_, place =  _decode_placement(first)
		if match:
			place = match.group(1)
		return info, place
	except Exception as exc:
		if DEBUG:
			print('_decode_placement, Error: {}'.format(exc))
			traceback.print_exc()
			pdb.set_trace()
		else:
			print('{}{}_decode_placement, string is a{}{}{}'
				  .format(exc, NL, type(string), NL, string))

def _unquote( string ):
	if string[ 0 ] in QUOTES and string[ -1 ] == string[ 0 ]:
		string = string[ 1:-1 ]
	return string

CurrentEnt = None
def _proc_entities( file, pl ):
	global CurrentEnt

	try:
		ordered = [(_decode_placement(cdent), cdent) for cdent in pl.keys()]
		# restrict sort to placement only, so keys are in original order
		ordered.sort(key=lambda x: 0 if x[0][1] is None else int(x[0][1]))
		for dec_pl, cdent in ordered:
			ent, place = dec_pl
			CurrentEnt = ent
			if ent not in Registered_ents.keys():
				Registered_ents[ ent ] = Entity( ent )
			entity = Registered_ents[ ent ]
			if file not in SourceFileNames:
				SourceFileNames.append( file )
			origID = entity.add_origin( SourceFileNames.index( file ), place )
			for cdprop in _sort_dict_keys( pl[ cdent ] ):
				prop, pplace = _decode_placement( cdprop )
				if len( Keys_to_output ) == 0 or prop in Keys_to_output:
					# len == 0 => process all attrs
					_proc_entity_key( pl, cdent, cdprop, ent, origID )
				else:
					# _log_error('_proc_entities, found attribute ({}) with no directive'.format(prop))
					pass
		CurrentEnt = None
	except Exception as exc:
		if DEBUG:
			print('_proc_entities, Error: {}'.format(exc))
			traceback.print_exc()
			pdb.set_trace()
		else:
			print( '{}\n_proc_entities, file is {}'.format( exc, file ))

def _add_entity_key( pl, cdent, cdprop, origID ):
	# noinspection PyUnusedLocal
	ent = eplace = prop = pplace = cdvalue = value = propInst = None
	try:
		ent, eplace = _decode_placement( cdent )
		prop, pplace = _decode_placement( cdprop )
		cdvalue = pl[ cdent ] if cdprop is None else pl[ cdent ][ cdprop ]
		value, vplace = _decode_placement( cdvalue )
		if prop in Keys_to_rename:
			prop = Keys_to_rename[ prop ]
			key = _encode_placement( prop, pplace )
		else:
			key = cdprop
		if cdprop is None:
			propInst = Property( ent, eplace, value, vplace, origID )
		else:
			propInst = Property( prop, pplace, value, vplace, origID )
		reg_ent = Registered_ents[ ent ]
		reg_ent.add_item( prop, propInst )
		if key is not None and prop in Keys_to_insert.keys():
			iplace = 0 if pplace is None else int( pplace )
			for ikey, ivalue in Keys_to_insert[ prop ].items():
				# must ensure a value is inserted for each origID,
				# else lists won't align on output
				if all(ikey not in key for key in pl[cdent].keys()): # is present in cdent
					iproperty = Property(ikey, iplace - 2, ivalue,
										  iplace - 1, origID, True)
					iproperty.plist.append(str(ivalue) + ';')
					reg_ent.add_item(ikey, iproperty)
		return prop, pplace, cdvalue, value, propInst
	except Exception as exc:
		if DEBUG:
			print('_add_entity_key, Error: {}'.format(exc))
			traceback.print_exc()
			pdb.set_trace()
		else:
			print('{}\n_add_entity_key, ent is {}, cdprop is {}'.format(
					exc, ent, cdprop))

def _is_float( string ):
	try:
		float( string )
	except ValueError:
		return False
	else:
		return True
		
def _proc_entity_key( pl, cdent, cdprop, ent, origID ):
	try:
		prop, pplace, cdvalue, value, propInst = _add_entity_key( pl, cdent, cdprop, origID )
		vlist = [] # list of tokens for value's output
		compact = 0 # default is now, N > 0 delay for N levels
		if prop in Keys_to_compact:
			compact = 1 # new style definition require special formatting (less whitespace)
			if prop == 'subentities': # old vs new style: list of str vs list of dict
				if len( cdvalue ) > 0:
					compact = 1 if isinstance( cdvalue[0], str ) else 2
		if isinstance( cdvalue, list ):
			idx_to_del = []
			for i in range( len( cdvalue ) ):
				dct = cdvalue[ i ]
				if RemoveFlasherSubEnts and prop == 'subentities':
					if isinstance( dct, str ):
						flasher, _ = _decode_placement( dct )
						if flasher.startswith( '*FLASHER*' ): # armoury_fieldMine
							idx_to_del.append( i )
					elif isinstance( dct, dict ):
						for encKey, encVal in dct.items():
							dk, _ = _decode_placement( encKey )
							dv, _ = _decode_placement( encVal )
							if dk == 'type' and dv == 'flasher':
								idx_to_del.append( i )
			idx_to_del.reverse()
			for i in idx_to_del:
				cdvalue.pop( i )
			if len( cdvalue ) > 0:
				write_list( vlist, cdvalue, 2, compact )
			elif not propInst.inserted:
				del Registered_ents[ ent ].properties[ prop ]
		elif isinstance( cdvalue, dict ):
			if prop in Subkeys_to_insert_if_empty.keys():
				subs_to_insert = Subkeys_to_insert_if_empty[ prop ]
				for encValKey, encValValue in value.items(): # decode subkeys so can test if empty
					if not isinstance( encValValue, dict ): continue
					valKey, _ = _decode_placement( encValKey )
					subkeys = {}
					for encVk, encVv in encValValue.items():
						vk, _ = _decode_placement( encVk )
						vv, _ = _decode_placement( encVv )
						subkeys[ vk ] = vv
					existingKeys = subkeys.keys()
					if any( k in existingKeys for k in subs_to_insert.keys() ): # see if in use
						continue # not empty
					for keyToAdd, valueToAdd in subs_to_insert.items():
						value[ encValKey ][ keyToAdd ] = valueToAdd
			write_dict( vlist, cdvalue, 2, compact )
		elif isinstance( cdvalue, bool ):
			vlist.extend( [ 'yes' if value else 'no', ';' ] )
		elif isinstance( value, float ) or isinstance( value, int ):
			vlist.extend( [ value, ';' ] )
		elif value.isdigit():
		# elif value.isdecimal(): # Python3 only (there are 4 other instances)
			vlist.extend( [ int( value ), ';' ] )
		elif _is_float( value ):
			vlist.extend( [ float( value ), ';' ] )
		else:
			if len( value ) == 0:
				quote = '"' # allow for empty strings
			else:
				quote = '"'  if NONWORD.search( value ) else ''
			vlist.append( '{}{}{};'.format( quote, value, quote ) )
		propInst.plist.extend( vlist )
		propInst.length = sum( len( str( pl ) ) for pl in vlist )
	except Exception as exc:
		if DEBUG:
			print('_proc_entity_key, Error: {}'.format(exc))
			traceback.print_exc()
			pdb.set_trace()
		else:
			print('{}\n_proc_entity_key, ent is {}, cdprop is {}'.format(
					exc, ent, cdprop))
		
		
## NextStep format ############################################################


# write_dict() & write_list() adapted from work by Cdr. Jettison
# http://aegidian.org/bb/viewtopic.php?f=2&t=10039&start=39

NONWORD = re.compile(r'\W') # if keys contain this pattern, they'll be
							# put in quotes (the keys on the outmost
							# level are always quoted)
QUOTEDNUMBER = re.compile( r'''
				^(?P<qt>[''' + QUOTES + r''']?)         # opening quote (maybe)
				(
						[-+]?                           # sign (maybe)
					(\d+ (?!/)                          # number (with no slash)
						(\.\d*)?)                       # a decimal part (maybe)
						( [eE] [-+]? \d+ )?             # exponent (maybe)
				  | ( (\d+)\s*/\s*(\d+) )               # or fraction
				)
				(?(qt) (?P=qt) | $ )                    # matching quote (if present)
				''', re.VERBOSE ) # strict pattern as used to omit quotes in output

def write_str( value, fmt, plist ):
	if len( value ) == 0 or any( c in value for c in WHITESPACE_CHAR ):
		quote = '"'
	else:
		qtnum = QUOTEDNUMBER.match( value )
		if qtnum:
			quote = ''
			value = qtnum.group( 2 ) # group 1 is the quote
		else:
			quote = '"' if NONWORD.search( value ) else ''
	plist.append( fmt.format( quote, value, quote ) )
			   
def _sort_dict_keys( aDict ):
	keys = aDict.keys()
	order = []
	try:
		for key in keys:
			info, place = _decode_placement( key )
			order.append( [ 0 if place is None else int( place ), key ] )
		order.sort( key = itemgetter( 0 ) )  # restrict sort to placement only
	except Exception as exc:
		if DEBUG:
			print('_sort_dict_keys, Error: {}'.format(exc))
			traceback.print_exc()
			pdb.set_trace()
		else:
			print( '{}\n_sort_dict_keys, keys is {}'.format( exc, keys ))
	return list( item[ 1 ] for item in order ) # ordered list of keys

def _sort_list_items( array ):

	def get_list_placement( x ):
		_, pl = _decode_placement( x )
		return pl

	def get_dict_placement( x ):
		k, v = x
		_, pl = _decode_placement( k )
		return pl

	order = []
	for index in range( len( array ) ):
		item = array[index]
		if isinstance( item, dict ):
			places = list( map( get_dict_placement, item.items() ) )
			place = min( 1e10 if place is None else place for place in places )
		elif isinstance( item, list ):
			places = list( map( get_list_placement, item ) )
			place = min( 1e10 if place is None else place for place in places )
		else:
			_, place = _decode_placement( item )
		if place is None: # no placement (XML source), fit into order
			placement = 0 if len( order ) == 0 else order[ -1 ][ 0 ] + 1 # incr placement of last entry
		else:
			placement = int( place )
		order.append( [ placement, index ] )
	order.sort( key = itemgetter( 0 ) )  # restrict sort to placement only         
	return list( item[ 1 ] for item in order )  
							
def line_empty( lst, endings = '' ):
	"""test line (str or list of str) is blank
	   or ends in member(s) of 'endings'"""
	try:
		start = _find_line_start( lst, len( lst ) )
		if start is None:
			start = 0 # no NL char
		trim = len( endings ) == 0
		i = len( lst )
		while i > start:
			i -= 1
			word = lst[ i ]
			strpped = word.strip() if trim else word.rstrip()
			if NL in word and strpped == '':
				return True
			if strpped == '':
				continue
			if trim:
				return False
			if any( strpped.endswith( c ) for c in endings ):
				return True
			return False
		return True
	except Exception as exc:
		if DEBUG:
			print('line_empty, Error: {}'.format(exc))
			traceback.print_exc()
			pdb.set_trace()
		else:
			print( '{}\nline_empty, item, type {}, is "{}"'.format(
					exc, type(item), lst[index]))
	
def write_dict( plist, dct, lvl, compact = 0, inlist = False ):
	"""Writes dictionary into object plist. Recursion as needed"""

	value = None # init for exception print()
	try:
		same_line = compact == 0 and lvl > 3
		if not same_line and (len(plist) == 0 or not line_empty(plist, '({,')):
			plist.extend([ NL, TAB * lvl ])
		plist.append('{')
		order = _sort_dict_keys(dct)
		for cdkey in order:
			key, place = _decode_placement(cdkey)
			cdvalue = dct[cdkey]
			value, _ = _decode_placement(cdvalue)
			if not same_line and not line_empty(plist):
				plist.extend([NL, TAB * (lvl + 1)])
			quote = '"' if NONWORD.search(key) else ''
			plist.extend(['{}{}{}'.format(quote, key, quote ), '='])
			if isinstance(cdvalue, list):
				if lvl == 3 and key in Subkeys_to_compact:
					nextcompact = 1
				else:
					nextcompact = 0 if compact <= 1 else compact - 1
				write_list(plist, value, lvl + 1, nextcompact, False)
			elif isinstance( cdvalue, dict ):
				if lvl == 3 and key in Subkeys_to_compact: 
					nextcompact = 1
				else:
					nextcompact = 0 if compact <= 1 else compact - 1
				write_dict(plist, value, lvl + 1, nextcompact, False)
			elif isinstance(cdvalue, bool):
				# xml decoder does not supply placement, so they're
				# place values are 0 (place==0 for OpenStep is always the
				# files first '{')
				bool_str = boolStr(cdvalue, OpenStep=place>0)
				plist.extend([bool_str, ';'])
			elif isinstance(value, (float, int)):
				plist.extend([value, ';'])
			elif value.isdigit():
				plist.extend([int( value ), ';'])
			elif _is_float(value):
				plist.extend([float(value), ';'])
			else: # isinstance( value, str )
				write_str(value, '{}{}{};', plist)
		if not same_line:
			plist.extend( [ NL, TAB * lvl ] )
		plist.append( '}' )
		if not inlist: plist.append( ';' )
	except Exception as exc:
		if DEBUG:
			print('write_dict, Error: {}'.format(exc))
			traceback.print_exc()
			pdb.set_trace()
		else:
			print( '{}\nwrite_dict, value, type {}, is "{}"'.format( exc, type( value ), value ))
	return False

def write_list( plist, lst, lvl, compact = 0, inlist = False ):
	"""Writes list as array into object plist. Recursion as needed"""

	value = None
	try:
		same_line = compact == 0 and lvl > 2
		if not same_line and (len( plist ) == 0 or not line_empty( plist, '({,' )):
			plist.extend( [ NL, TAB * lvl ] )
		plist.append( '(' )
		first = True
		order = _sort_list_items( lst )
		for index in order:
			cdvalue = lst[ index ]
			value, place = _decode_placement( cdvalue )
			if not first: plist.append( ',' )
			first = False
			if not same_line and not line_empty( plist ):
				plist.extend( [ NL, TAB * (lvl + 1) ] )
			if isinstance( cdvalue, list ):
				write_list( plist, cdvalue, lvl + 1, 0 if compact <= 1 else compact - 1, True )
			elif isinstance( cdvalue, dict ):
				# write_dict returns True for comments to suppress commas
				first = write_dict( plist, cdvalue, lvl + 1, 0 if compact <= 1 else compact - 1, True )
			elif isinstance( cdvalue, bool ):
				# xml decoder does not supply placement, so they're
				# place values are 0 (place==0 for OpenStep is always the
				# files first '{')
				bool_str = boolStr(cdvalue, OpenStep=place>0)
				plist.append(bool_str)
			elif isinstance( value, float ) or isinstance( value, int ):
				plist.append( value )
			elif value.isdigit():
				plist.append( int( value ) )
			elif _is_float( value ):
				plist.append( float( value ) )
			else:
				write_str( value, '{}{}{}', plist )
		if not same_line:
			plist.extend( [ NL, TAB * lvl ] )
		plist.append( ')' )
		if not inlist: plist.append( ';' )
	except Exception as exc:
		if DEBUG:
			print('write_list, Error: {}'.format(exc))
			traceback.print_exc()
			pdb.set_trace()
		else:
			print( '{}\nwrite_list, value, type {}, is "{}"'.format( exc, type( value ), value ))

	
## OpenStep parser ############################################################

'''
MIT License
===========

Copyright (c) 2017, Samuel Bishop
Copyright (c) 2014, Fireteam Ltd.
See AUTHORS for more details.

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

- 'decoder.py' & 'scanner.py' included and (heavily) modified for oolite
'''
STRIP_PLACE_RE = re.compile(r'''
			\# \d+ \#                            # placement number
			   ( .*? )
			\#\#
			''', re.VERBOSE|re.DOTALL|re.MULTILINE)
   
def linecol( doc, pos ):
	lineno = doc.count( NL, 0, pos ) + 1
	if lineno == 1:
		colno = pos
	else:
		colno = pos - doc.rindex( NL, 0, pos )
	return lineno, colno

def errmsg( msg, doc, pos, index = None ):
	decmsg = STRIP_PLACE_RE.sub( r'\1', msg )
	lineno, colno = linecol( doc, pos )
	if index is None:
		fmt = 'line {0}, column {1} (char {2}): {3}'
		error = fmt.format( lineno, colno, pos, decmsg )
	else:
		endlineno, endcolno = linecol( doc, index )
		fmt = 'line {0}, column {1} - line {2}, column {3} (char {4} - {5}); {6}'
		error = fmt.format( lineno, colno, endlineno, endcolno, pos, index, decmsg )
	error += errcontext( doc, pos )
	return error
	
def errcontext( doc, pos ):
	start = 0 if pos < 40 else pos - 40
	errnl = NL + '    ... '
	errpref = doc[start:pos].replace(NL, '').replace(TAB, '  ').replace('\r', '')
	pointer = NL + ' ' * (len(errnl) + len(errpref) - 1) + '^'
	errsuff = doc[pos:pos + 40].replace(NL, '').replace(TAB, '  ').replace('\r', '')
	return errnl + errpref + errsuff + ' ...' + pointer

#STRINGCHUNK = re.compile(r'''(.*?)(["'\\\x00-\x1f])''', re.VERBOSE|re.DOTALL)
QUOTES_SUPPORTED = [] 
# make regex for each quote char to ensure starting matches ending and others allowed in string
#   "(?:\\.|[^"\\])*"   from StackOverflow
	
for qt in QUOTES:
	# QUOTES_SUPPORTED.append( [ qt, re.compile( r'(.*?)([' + qt + '\\\x00-\x1f])' ) ] )
	QUOTES_SUPPORTED.append( [ qt, re.compile( r'((?:\\.|[^' + qt + r'\\])*)([' + qt + r'\x00-\x1f\x7f])' ) ] )

def PListString( string, index, encoding = "utf-8" ):
	"""Scan 'string' for a plist quoted string. 'index' is of the
	character in 'string' after the quote that started the plist string.
	Unescapes all valid plist string escape sequences and raises ValueError
	on attempt to decode an invalid string.

	Returns a tuple of the decoded string and the index of the character in string
	after the end quote."""
	
	chunks = []
	begin = index - 1
	initiator = string[ index - 1 ] # quote char that starts a quoted string
	_match = None
	for quote, regex in QUOTES_SUPPORTED:
		if initiator == quote:
			_match = regex.match
			break
	if _match is None:
		raise ValueError( errmsg( "Unsupported quote char for string starting at", string, begin ) )       
	while True:
		chunk = _match( string, index )
		if chunk is None:
			raise ValueError( errmsg( "Unterminated string starting at", string, begin ) )
		index = chunk.end()
		content, terminator = chunk.groups() # each QUOTES_SUPPORTED has only 2 groups
		# Content is contains zero or more unescaped string characters
		if content:
			if not isinstance( content, str ):
				content = content.encode( encoding )
				# content = str( content, encoding )	# works in Python3
			chunks.append( content )
		# Terminator is the end of string, a literal control character,
		# or a backslash denoting that an escape sequence follows
		if terminator in QUOTES:
			break
		elif terminator != '\\':
			if string[ index - 1 ] == TAB: # oolite tolerates tabs in strings
				chunks.append(TAB)
				continue
			else:
				msg = "Invalid control character {0!r} at".format( terminator )
				raise ValueError( errmsg( msg, string, index ) )
		try:
			esc = string[ index ]
		except IndexError:
			raise ValueError( 
				errmsg( "Unterminated string starting at", string, begin ) )
		# If not a unicode escape sequence, must be in the lookup table
		if esc != 'U':
			msg = "Invalid \\escape: " + repr( esc )
			raise ValueError( errmsg( msg, string, index ) )
		else:
			# Unicode escape sequence
			esc = string[ index + 1:index + 5 ]
			next_end = index + 5
			if len( esc) != 4:
				msg = "Invalid \\UXXXX escape"
				raise ValueError( errmsg( msg, string, index ) )
			uni = int( esc, 16 )
			# Check for surrogate pair on UCS-4 systems
			if 0xd800 <= uni <= 0xdbff and sys.maxunicode > 65535:
				msg = "Invalid \\UXXXX\\UXXXX surrogate pair"
				if not string[ index + 5:index + 7 ] == '\\u':
					raise ValueError( errmsg( msg, string, index ) )
				esc2 = string[ index + 7:index + 11 ]
				if len( esc2) != 4:
					raise ValueError( errmsg( msg, string, index ) )
				uni2 = int( esc2, 16 )
				uni = 0x10000 + ( ( (uni - 0xd800) << 10 ) | (uni2 - 0xdc00) )
				next_end += 6
			char = chr( uni )
			index = next_end
		# Append the unescaped character
		chunks.append( char )
	
	return ''.join( chunks ), index

WHITESPACE_RE = re.compile(r'[ \t\n\r]*', re.VERBOSE|re.MULTILINE|re.DOTALL)
WHITESPACE_CHAR = ' \t\n\r'

def _skip_WS( string, index, _WSmatch = WHITESPACE_RE.match,
			  _WSstr = WHITESPACE_CHAR ):
	started = index # save index before skip for correct error msgs
	nextchar = ''
	try:
		nextchar = string[ index ]
		if nextchar in _WSstr:
			index = _WSmatch( string, index ).end()
			nextchar = string[ index ]
		if nextchar == '/':
			return PListComment( string, index + 1 )
	except IndexError:
		pass # found none
	return nextchar, index, started

CLOSECOMMENT = re.compile(r'.*?(\*/)', re.VERBOSE|re.MULTILINE|re.DOTALL)
MATCH2EOL = r'(.*?\n)'
EOLCOMMENT = re.compile('(' + ORIGIN_MARKER + ')?' + MATCH2EOL, re.VERBOSE)

def PListComment( string, index ):
	""" comments can appear anywhere, so this gets called often,
		every time we 'scan' or skip whitespace
	"""

	nextchar = string[ index ]
	eol_style = nextchar == '/'
	find_close = nextchar == '*'
	try: # must gobble up ALL consecutive comments, even a mixture (but not nested!)
		if eol_style: # need to inspect ea for ORIGIN_MARKER
			match = EOLCOMMENT.match( string, index + 1 )
			if match is not None: 
				index = match.end() # points at eol
			else:
				raise ValueError( errmsg( "Invalid comment - no eol??", string, index ) ) 
		elif find_close:
			match = CLOSECOMMENT.match( string, index + 1 )
			if match is not None: 
				index = match.end()
			else:
				raise ValueError( errmsg( "Invalid comment - no closing */", string, index ) )
	except IndexError:
		raise StopIteration
	return _skip_WS( string, index )

DictLevel = 0
def PListDictionary(string, index):
	global DictLevel

	nextchar, index, _ = _skip_WS(string, index)
	pairs = []
	# Trivial empty object
	if nextchar == '}':
		return pairs, index + 1
	DictLevel += 1
	while True:
		key_start = index
		try: # fetching the key
			key, index = PL_scan_name(string, index)
		except StopIteration:
			raise ValueError(errmsg('Expecting property name', string, index))

		if LogUnquotedDash and '-' in key and string[ key_start ] not in QUOTES:
			key = '"' + key + '"'
			msg = '(non-fatal) Unquoted ship key '
			msg += "{} containing a '-'".format( key )
			_log_error(errmsg(msg, string, index))
		ekey = _encode_placement( key, index ) 
		
		nextchar, index, started = _skip_WS( string, index )
		if nextchar != '=':
			raise ValueError(errmsg('Expecting = delimiter', string, started))
		index += 1 # step past '='

		nextchar, index, started = _skip_WS( string, index )
		try: # fetching the key's value
			value, index = PL_scan_once( string, index )
		except StopIteration:
			delim = -1
			if index < len( string ):
				# re failed, assume "" (oxp is verified, right?)
				semic = string.find( ';', index )
				comma = string.find( ',', index )
				if semic != -1 and semic < comma: delim = semic
				elif comma != -1 and comma < semic: delim = comma
			if delim != -1:
				msg = '(non-fatal) Found no value for {}, using ""'.format(key)
				_log_error(errmsg(msg, string, started))
				value = '""'
				index = delim
			else:
				raise ValueError(errmsg('Expecting dictionary', string, index))
		evalue = value       
		if type( value ) not in [ dict, list ]: # encode for proper placement
			evalue = _encode_placement( value, index )
		pairs.append( (ekey, evalue) )
		nextchar, index, started = _skip_WS( string, index )
		if nextchar == ';' or \
			( nextchar == ',' and type( value ) in [ 'list', 'dict' ] ):
			index += 1# step past semi-colon/comma
		elif nextchar in ')}': # missing final semi-colon (oolite tolerates)
			if LogMissingSemiColon: # will be added on output
				_log_error(errmsg('(non-fatal) missing ";"', string, started))
		else:
			raise ValueError(errmsg('Expecting ; or , delimiter',
									string, index - 1))

		nextchar, index, _ = _skip_WS( string, index )
		if nextchar == '}':
			index += 1 # step past closing brace
			break

	pairs = dict( pairs )
	DictLevel -= 1
	return pairs, index

def PListArray( string, index ):
	nextchar, index, _ = _skip_WS( string, index )
	values = []
	# Look-ahead for trivial empty array
	if nextchar == ')':
		return values, index + 1
	while True:
		try: # fetching an array item
			value, index = PL_scan_once( string, index )
		except StopIteration:
			raise ValueError( errmsg( "Expecting object", string, index ) )
		
		if type( value ) not in [ dict, list ]: # encode for proper placement
			value = _encode_placement( value, index )
		values.append( value )

		nextchar, index, _ = _skip_WS( string, index )
		if nextchar == ')':
			index += 1 # step past closing parenthesis
			break
		elif nextchar != ',':
			raise ValueError( errmsg( "Expecting , delimiter", string, index ))
		index += 1# step past comma

		nextchar, index, _ = _skip_WS( string, index )
		if nextchar == ')': # trailing commas are allowed
			index += 1 # step past closing parenthesis
			break

	return values, index

def PListDecode( string, index = 0 ):
	"""Decode a plist document from ``string`` (a ``str`` or ``unicode``
	beginning with a plist document) and return a 2-tuple of the Python
	representation and the index in ``string`` where the document ended.

	This can be used to decode a plist document from a string that may
	have extraneous data at the end.
	"""

	try:
		_, index, _ = _skip_WS(string, index)
		obj, index = PL_scan_once(string, index)
	except MatchFailError as err:
		msg = 'PListDecode, No plist object could be decoded '
		msg += '(MatchFailError at "{}").'.format(err.nextchar)
		_log_error(errmsg(msg, string, err.index))
		return None, 0
	except StopIteration:
		msg = 'PListDecode, No plist object could be decoded (StopIteration).'
		_log_error(errmsg(msg, string, index))
		return None, 0
	except ValueError as err: # catch-all for parse errors
		_log_error('PListDecode, {}'.format(err))
		return None, 0
	return obj, index

ALPHANUMERIC = r'''(  
						( \d+ \s* / \s* \d+ )   	# a fraction
					  | (                  			# or number
						  [-+]?                     # sign (maybe)
						  ( \d+ (?!/)               # digits (with no slash)
							[.]?\d* | \.\d+ )       # decimal (maybe)
						  ( [eE] [-+]? \d+ )?       # exponent (maybe)
						)                 
					  | ( [a-zA-Z0-9\-\._]+ )   # or an identifier, filename, even with dashes
					)'''

ALPHANUMERIC_RE =  re.compile( ALPHANUMERIC, re.VERBOSE | re.DOTALL | re.MULTILINE )

# see http://code.google.com/p/networkpx/wiki/PlistSpec#Text        #### in scrapbook
def PL_scan_once( string, index ):
	if index >= len( string ): 
		raise StopIteration
	nextchar = string[ index ]
	if nextchar in QUOTES: 
		obj, index = PListString( string, index + 1 )
		return obj, index
	elif nextchar == '{': 
		obj, index = PListDictionary( string, index + 1 )
		return obj, index
	elif nextchar == '(': 
		obj, index = PListArray( string, index + 1 )
		return obj, index
	elif nextchar == '/': 
		nextchar, index, _ = PListComment( string, index + 1 )
		obj, index = PL_scan_once( string, index )
		return obj, index

	match = ALPHANUMERIC_RE.match( string, index )
	if match is not None: 
		return match.group( 0 ), match.end()
	else: 
		raise MatchFailError( index, string[ index ] )

def PL_scan_name( string, index ):
	if index >= len( string ): 
		raise StopIteration
	nextchar = string[ index ]
	if nextchar in QUOTES:
		obj, index = PListString( string, index + 1 )
		return obj, index
	elif nextchar == '/':
		nextchar, index, _ = PListComment( string, index + 1 )
		obj, index = PL_scan_name( string, index )
		return obj, index

	match = ALPHANUMERIC_RE.match( string, index )
	if match is not None: 
		return match.group( 0 ), match.end()
	else: 
		raise MatchFailError( index, string[ index ] )

	
## .plist file ################################################################


def _find_line_start( plist, index ):
	if len( plist ) == 0: return 0
	return _find_these_chars( plist, index, NL, -1, skip_content=True )

### no longer used ###
def _endswith_NL( plist, index = None ):
	if index is None:
		index = len( plist )
	return _find_these_chars( plist, index, NL, -1, skip_content=False )

def _startswith_NL( plist, index = -1 ):
	return _find_these_chars( plist, index, NL, 1, skip_content=False )
	
def _find_open_dict( plist, index ):
	return _find_these_chars( plist, index, '{', -1, skip_content=False )

def _find_close_dict( plist, index ):
	return _find_these_chars( plist, index, '}', 1, skip_content=False )
	
def _find_open_list( plist, index = None ):
	if index is None:
		index = len( plist )
	return _find_these_chars( plist, index, '(', -1, skip_content=False )

def _find_close_list( plist, index ):
	return _find_these_chars( plist, index, ')', 1, skip_content=False )

def _find_cmt_prefix( plist, index ):
	line_start = _find_these_chars( plist, index, NL, -1, skip_content=False )
	if line_start is not None: # nothing but whitespace
		return line_start
	found = _find_these_chars( plist, index, '', -1, skip_content=False )
	return None if found is None else found + 1

def _find_cmt_suffix( plist, index, eol_style = False ):
	if eol_style: # locate eol or bail
		line_end = _find_these_chars( plist, index, NL, -1, skip_content=False )
		if line_end is not None: # nothing but whitespace
			return line_end
		return index
	# find_close */: preserve trailing whitespace but not eol
	line_end = _find_these_chars( plist, index - 1, NL, 1, skip_content=False )
	if line_end is not None: # nothing but whitespace
		return line_end
	found = _find_these_chars( plist, index, '', 1, skip_content=False )
	return None if found is None else found - 1
### end no longer used ###

def _find_these_chars( plist, index, chars, direction, skip_content ):
	"""Search for any of 'chars' in a (list of) string(s) ('plist')
	   starting at 'index' stepping in 'direction' (+/-N),
	   perhaps skipping over content. No chars => any non-WS.
	   Return None if fail, index of string in plist
	   that contains a member of 'chars'."""

	token = None
	try:

		limits = range(index, len(plist)) if direction > 0 else range(0, index)
		trim = len(chars) == 0
		if trim:
			skip_content = False # both True is nonsense
		offset = index
		while True:
			offset += direction
			if offset not in limits: break # exceeded boundary
			token = plist[offset]
			if not isinstance(token, str): # found a numnber
				if not skip_content:
					return offset if trim else None
				continue
			if trim: # chars = '' means looking for any visible char
				if token.strip(WHITESPACE_CHAR) != '':
					return offset
				continue
			elif skip_content:
				if any(c in chars for c in token):
					return offset
				continue
			# search each character
			toklen = len(token)
			idx = toklen if direction < 0 else -direction
			while True:
				idx += direction
				if idx < 0 or idx >= toklen: break # exceeded limit
				tk = token[idx]
				if trim and tk not in WHITESPACE_CHAR:
					return offset # any visible char
				elif tk in chars:
					return offset
				elif not skip_content and tk not in WHITESPACE_CHAR: 
					return offset if trim else None
		return None

	except Exception as exc:
		if DEBUG:
			print('_find_these_chars, Error: {}'.format(exc))
			traceback.print_exc()
			pdb.set_trace()
		else:
			print('{}\n_find_these_chars, index = {}, token = {}'.format(
					exc, index, token))

def base_in_file(liked_prop):	# liked_prop is a list of Property from a like_effect
	"""walk back through chain of like_effect Property's to find base/template
	   If not found in file, caller must retain 'is_external_dependency' prop"""

	curr_file = max(prop.origID for prop in liked_prop)
	while True:
		liked_model = max(liked_prop, key=lambda x: x.origID).value
		if liked_model in Registered_ents:
			liked_prop = Registered_ents[liked_model].properties.get('like_effect', None)
			if liked_prop:
				liked_file = max(prop.origID for prop in liked_prop)
				if liked_file != curr_file:
					return False # likes continue beyond file, keep is_external_dependency
			else: # found base model
				liked_file = max(Registered_ents[liked_model].origin)
				return liked_file == curr_file # base model in same file?
				# liked_file = max(prop.origID for prop in
				# 				 Registered_ents[liked_model].properties.get('model', None))
				# return liked_file is not None and liked_file == curr_file # base model in same file?
		else:
			return False # base model not loaded

def same_as_liked(key, liked_prop, latest):
	"""follow the like_effect chain to get the value of key
	   Return bool indicating if values are equal"""

	while True:
		# liked_prop is a list of Property from a like_effect
		liked_model = max(liked_prop, key=lambda x: x.origID).value
		if liked_model in Registered_ents:
			liked_properties = Registered_ents[liked_model].properties
			liked_key = liked_properties.get(key, None)
			if liked_key:
				liked_value = max(liked_key, key=lambda x: x.origID)
				return _equal_values(liked_value.value, latest.value)
			# check further back in chain
			liked_prop = liked_properties.get('like_effect', None)
			if not liked_prop: # at end of like chain
				return False
		else: # key never defined before
			return False

def _render_values(name, simpleton):
	ent = Registered_ents[name]
	plist = []
	try:

		plist.extend([ NL + TAB, '"' + ent.name + '"', '=' ]) # identifiers are always quoted
		if not simpleton:
			plist.extend([ NL + TAB, '{' ])
		#  sort by property name
		ordered = sorted(ent.properties.items(), key=lambda x: x[0])

		# use property values with the highest origID (latest file to define/redefine ent)
		# - origID may differ among properties when using like_ship but earlier ones
		#   have been overridden (they're saved for error msgs), so use that with
		#   the highest origID

		# use latest one
		last_origID = max(pp.origID for key, props in ordered for pp in props)
		# like_effect is only in play if in latest version
		like_effect = [pp for key, props in ordered for pp in props
				 			if key == 'like_effect' and last_origID == pp.origID]
		for key, props in ordered: # props is a list of Property instances
			latest = max(props, key=lambda x: x.origID)
			if not like_effect and latest.origID != last_origID:
				continue # not relevant since key was overriden
			if like_effect and key not in Keys_not_suppressed:
				if same_as_liked(key, like_effect, latest):
					continue # value already exists in liked model
			if key == 'is_external_dependency' and latest.inserted:
				# not all are needed; these are inserted for all like_ship but only
				#   now can we know that the reference is to an ent in the same file
				# only suppress this key if the reference chain ends in a model
				#   that's defined in this file
				if base_in_file(like_effect): # don't need this key
					continue

			if not simpleton:
				plist.extend([ NL + TAB + TAB, key, '=' ])
			plist.extend(latest.plist)

		if not simpleton:
			plist.extend([NL + TAB, '};', NL + NL])
		return plist

	except Exception as exc:
		if DEBUG:
			print('_render_values, Error: {}'.format(exc))
			traceback.print_exc()
			pdb.set_trace()
		else:
			print( '{}\n_render_values, plist = {}'.format( exc, plist ))

def merge_prop_dicts(props, plist):
	""" output latest version of values for each unique identifier"""
	# when property values are themselves dicts, we inspect said dict's keys
	#   and merge latest of each unique identifier (eg. 'materials' has list of Property's)
	unique = {}
	for prop in props:
		for k, v in prop.value.items():
			info, place = _decode_placement(k)
			if info not in unique or prop.origID > unique[info].origID:
				# prop's .plist has all k's, treat as a unit
				if prop not in unique.values():
					unique[info] = prop
	length = len(unique)
	if length == 1:
		working = list(unique.values())[0]
		plist.extend(working.plist)
	else:  # merge disparate definitions into a unified whole
		count = 1
		for uni in unique.values():
			working = uni.plist[:]
			if count > 1:
				# remove line with opening brace from
				#   [NL, TAB, TAB, '{', NL, TAB, TAB ...
				if '{' in working:
					del working[:working.index('{') + 1]
			if count < length:
				# remove line with closing brace from
				#   ... ';', NL, TAB, TAB, '}', ';'])
				working.reverse()
				if '}' in working:
					del working[:working.index('}') + 1]
				working.reverse()
				while len(working) and working[-1].isspace():
					working.pop()
			plist.extend(working)
			count += 1

def _select_origin():
	order = []
	try:
		def sourceRank(path):
			if OoFolders['appDir'] in path:
				return 3	# core
			if OoFolders['maoDir'] in path:
				return 2	# ManagedAddOns overrides core
			if OoFolders['aoDir'] in path:
				return 1	# AddOns overrides both

		for name, ent in Registered_ents.items():
			origins = ent.origin
			# we rank by order loaded, ie. which key of OoFolders
			sources = [(origins.index(fn), fn, sourceRank(SourceFileNames[fn]))
					   for fn in origins]
			# sort by load priority (ie. latest is best)
			sources.sort(key=itemgetter(2))
			order.append([name, ent.place, sources[0][1]])	# sources's first fn
			ent.output_oi = sources[0][0]  					# sources's first index

		# group by file, placement (leave ents in original order)
		return sorted(order, key=itemgetter(2, 1))
	except Exception as exc:
		if DEBUG:
			print('_select_origin, Error: {}'.format(exc))
			traceback.print_exc()
			pdb.set_trace()
		else:
			print( '{}\n_select_origin, order = {}'.format( exc, order ))

# Registered_ents: dict name: Entity( name )
# 	.place = []			# array of indices in source where entity definition starts
# 	.origin = []		# list of file source indices where it was found
# 	.properties = {}	# dict of list(Property), mapping keys to list of value(s), eg. model: [] of Property

def validate_uniforms():
	global CurrentFile, Last_logged_fn

	textureLen = 0
	try:
		for name, ent in Registered_ents.items():
			propKeys = ent.properties.keys()
			props = []
			if 'materials' in propKeys: props.append(ent.properties['materials'])
			if 'shaders' in propKeys:	props.append(ent.properties['shaders'])
			if len( props ) == 0: continue
			for prop in props: # each a list of Property objects
				for pObj in prop:
					pVal = pObj.value # a dict of dict
					if not isinstance( pVal, dict ): continue
					for encpropKey, shDict in pVal.items():
						if not isinstance( shDict, dict ): continue
						textureFns = []
						uniformRefs = []
						msg = INPUT_FN + ' line '
						noErrLen = len( msg )
						propKey, _ = _decode_placement( encpropKey )
						for encShKey, encShValue in shDict.items():
							shKey, _ = _decode_placement( encShKey )
							shValue, _ = _decode_placement( encShValue )
							if shKey not in ['textures', 'uniforms']:
								continue
							if shKey == 'textures' and isinstance( shValue, list ):
								for sValue in shValue:
									decValue, _ = _decode_placement(sValue)
									if isinstance( decValue, str ):
										textureFns.append( decValue )
									elif isinstance( decValue, dict ):
										for skey, sval in decValue.items():
											decSubKey, _ = _decode_placement(skey)
											decSubValue, _ = _decode_placement(sval)
											if decSubKey == 'name':
												textureFns.append(decSubValue)
									else:
										msg = 'shKey == textures, unknown type '
										msg += '{}, for decValue:{}{}'.format(
												type(decValue), NL, decValue)
										raise ValueError(msg)
								textureLen = len( textureFns )
							elif shKey == 'uniforms' and isinstance( shValue, dict ):
								for encUnKey, encUnVal in shValue.items():
									if not isinstance( encUnVal, dict ): continue
									unKey, _ = _decode_placement( encUnKey )
									decUnVal = {}
									# decode dictionary
									for encKey, encVal in encUnVal.items():
										decKey, _ = _decode_placement( encKey )
										decVal, _ = _decode_placement( encVal )
										decUnVal[ decKey ] = decVal
									decUnValKeys = decUnVal.keys()
									if 'type' in decUnValKeys \
											and decUnVal['type'] == 'texture' \
											and 'value' in decUnValKeys:
										textureRef = decUnVal[ 'value' ]
										if not textureRef.isdigit():
											msg += NL + '    for key "{}":'.format( name )
											msg += NL + '      Invalid reference '
											msg += 'value "{}" '.format(textureRef)
											msg += 'for uniform of {}'.format(propKey)
										else:
											textureRef = int( textureRef )
											uniformRefs.append( textureRef )
							else: # oops
								print('oops!   shKey =', shKey)
								print('oops! shValue =', shValue)
								oops = 'unknown key, value pair; '
								oops += 'encShKey "{}", '.format(encShKey)
								oops += 'encShValue:{}{}'.format(NL, encShValue)
								raise ValueError(oops)

						badrefs = list( ref for ref in uniformRefs if ref >= textureLen )
						if len( badrefs ) > 0:
							print( '! textureLen =', textureLen, 'textureRef =', uniformRefs )
							msg += NL + '    for key "{}":'.format( name ) + NL + '    '
							msg += '  Bad texture reference{} {} for uniform of "{}"'.format(
								   's' if len( badrefs ) > 1 else '', badrefs, propKey )
							msg += NL + '        textures list:' + NL
							count = 0
							for fn in textureFns:
								msg += '          ' + str( count ) + ': '
								if isinstance( fn, dict ) and 'name' in fn.keys():
									msg += fn[ 'name' ]
								elif isinstance( fn, str ):
									msg += fn
								else:
									msg += 'Unknown file!'
								msg += NL
								count += 1
						if len( msg ) > noErrLen:
							_ent_file_msg( ent.origin, name )  # calling for logging if dups
							CurrentFile = SourceFileNames[ ent.origin[ pObj.origID ] ]
							Last_logged_fn = None
							_log_error( msg )
	except Exception as exc:
		if DEBUG:
			print('validate_uniforms, Error: {}'.format(exc))
			traceback.print_exc()
			pdb.set_trace()
		else:
			print( '{}\nvalidate_uniforms, order = {}'.format( exc, order ))

def _gen_dup_lists( origins ):
	dup_props = dict()
	try:
		origins_ents = [ ent for ent in Registered_ents.keys()
						 if Registered_ents[ ent ].origin == origins ]
		for ent in origins_ents:
			reg_ent = Registered_ents[ ent ]
			props = reg_ent.properties
			if len( reg_ent.origin ) > 1:
				dup_props[ ent ] = []
			for prop, values in props.items():
				if len( values ) <= 1: continue
				if len( [ v for v in values if not v.inserted ] ) <= 1: continue
				if LogOnlyDifferentDups:
					output = values[ reg_ent.output_oi ]
					# OpenStep has placement info; vplace == 0 means is from XML
					outstr = _str_minimize(output.value, OpenStep=output.vplace > 0)
					different = False
					for v in values:
						if v.value == output.value: continue
						valstr = _str_minimize(v.value, OpenStep=v.vplace > 0)
						if valstr == outstr: continue
						different = True
						break
					if not different: continue
				if ent not in dup_props.keys():
					dup_props[ ent ] = []
				origp = prop
				if prop in Keys_to_rename.values():
					origp = [ k for k,v in Keys_to_rename.items()
							  if v == prop ][ 0 ]
				dup_props[ ent ].append( (prop, origp) )
			if LogOnlyDifferentDups and ent in dup_props.keys() \
					and len( dup_props[ ent ] ) == 0:
				del dup_props[ ent ]
	except Exception as exc:
		if DEBUG:
			print('_gen_dup_lists, Error: {}'.format(exc))
			traceback.print_exc()
			pdb.set_trace()
		else:
			print( '{}\n_gen_dup_lists, origins = {}'.format( exc, origins ))
	return dup_props # a dict of lists of tuples,  ent: [(prop,origp),...]

def _gen_log_header( origins, ent, dup_props ):
	try:
		leno = len( origins )
		lend = len( dup_props )
		if leno == 0: # ent only appears in collated file
			if lend > 0 :
				_ent_load_warning( ent )
			return
		if LogOnlyDifferentDups and sum( len( p ) for p in dup_props.values() ) == 0:
			return
		if leno > 1 or lend > 0: # have dups -> output file header
			_ent_file_msg( origins, ent, dup_props )
	except Exception as exc:
		if DEBUG:
			print('_gen_log_header, Error: {}'.format(exc))
			traceback.print_exc()
			pdb.set_trace()
		else:
			print( '{}\n_gen_log_header, origins = {}\nent = {}, dup_props: {}'
				   .format( exc, origins, ent, dup_props ))
	return dup_props # a dict of lists of tuples,  ent: [(prop,origp),...]
	  
def _simplify_list( lst ):
	if len( lst ) == 0: return ''
	simple = []
	try:
		for item in lst:
			decit, _ = _decode_placement( item )
			if isinstance( item, list ):
				simple.append( _simplify_list( item ) )
			elif isinstance( item, dict ):
				simple.append( _simplify_dict( item ) )
			else:
				simple.append( decit )
		simple.sort() # sort for comparison
	except Exception as exc:
		if DEBUG:
			print('_simplify_list, Error: {}'.format(exc))
			traceback.print_exc()
			pdb.set_trace()
		else:
			print( '{}{}_simplify_list, lst is a{}{}{}'
				   .format(exc, NL, type(lst), NL, lst))
	return '(' + ', '.join( simple ) + ')'
			 
def _simplify_dict( dct ):
	if len( dct ) == 0:
		return ''
	simple = []
	result = ''  # sort required so final strings can be compared
	try:
		for key, value in dct.items():
			deckey, _ = _decode_placement( key )
			decval, _ = _decode_placement( value )
			if isinstance( value, list ):
				simple.append( [ deckey, _simplify_list( value ) ] )
			elif isinstance( value, dict ):
				simple.append( [ deckey, _simplify_dict( value ) ] )
			else:
				simple.append( [ deckey, decval ] )
		simple = sorted( simple, key = lambda x: str( x[ 0 ] ) + str( x[ 1 ] ) )
		for key, value in simple:
			result += '{} = {};'.format( key, value )
	except Exception as exc:
		if DEBUG:
			print('_simplify_dict, Error: {}'.format(exc))
			traceback.print_exc()
			pdb.set_trace()
		else:
			print( '{}{}_simplify_dict, dct is a{}{}{}'
				   .format(exc, NL, type(dct), NL, dct))
	if result == '':
		return ''
	return '{' + result + '}'

def boolStr(value, OpenStep):
	check = value.lower()
	if OpenStep:
		if check in TRUEVALUES:
			return 'yes'
		elif check in FALSEVALUES:
			return 'no'
		raise ValueError('boolStr, "{}" is not a boolean, OpenStep: {}'
						 .format(value, OpenStep))
	else:
		return 'yes' if check else 'no'

def _str_minimize( string, translate = True, OpenStep = False ):
	min_str = string
	try:
		if string is None:
			return ''
		decstr, _ = _decode_placement( string )
		if isinstance( min_str, list ):
			min_str = _simplify_list( min_str )
		elif isinstance( min_str, dict ):
			min_str = _simplify_dict( min_str )
		elif isinstance( min_str, bool ) \
				or (OpenStep and decstr.lower() in BOOLEANS):
			min_str = boolStr(decstr, OpenStep)
		elif isinstance( decstr, float ) or isinstance( decstr, int ):
			min_str = str( decstr )
		elif decstr.isdigit():
			min_str = str( int( decstr ) )
		elif _is_float( decstr ):
			min_str = str( float( decstr ) )
		elif isinstance( min_str, str ):
			min_str = decstr.strip( QUOTES )
		else:
			raise 'a little hell'
		if isinstance( min_str, str ):
			if translate:
				min_str = min_str.translate( MakeSparse )
			while min_str.find( '  ' ) >= 0:
				min_str = min_str.replace( '  ', ' ' )
			min_str = min_str.strip()
	except Exception as exc:
		if DEBUG:
			print('_str_minimize, Error: {}'.format(exc))
			traceback.print_exc()
			pdb.set_trace()
		else:
			print( '{}{}_str_minimize, min_str is a{}{}{}'
				   .format(exc, NL, type(min_str), NL, min_str))
	return min_str

def _equal_values( new, curr ):
	if new == curr: return True
	newx = _str_minimize( new )
	currx = _str_minimize( curr )
	if newx == currx: return True
	return False

def _gen_eff_data(out_str):
	global Output_size

	Output_size += len(out_str)
	Output_size += out_str.count(NL)
	return out_str	# convience return


## main #######################################################################

def _create_eff_data():
	global  Output_size

	name = item = None  # for except print
	try:
		if len(Registered_ents) == 0:
			return
		input_files = []
		ordered = _select_origin()
		if AuditUniformTextures:
			validate_uniforms()
		output = ''
		for name, _, _ in ordered:
		# for name, place, file in ordered:
			ent = Registered_ents[name]
			simpkeys = ent.properties.keys()
			simpleton = True if len(simpkeys) == 1 and None in simpkeys else False
			origin = ent.origin
			if origin not in input_files:
				input_files.append(origin)
				if LogDuplicateKeys and not simpleton and len(origin) > 1:
					dup_props = _gen_dup_lists(origin)
					_gen_log_header(origin, name, dup_props)
					if len(dup_props) > 0:
						_ent_dup_error(dup_props)
			tokens = _render_values(name, simpleton)
			count = len(tokens)
			lastout_ws = False
			for num in range(count):  # format tokens with 'smart' spacing
				item = tokens[num]
				if item == '': continue
				if not isinstance(item, str):  # int or float
					skip_blank = lastout_ws
					append_blank = False
				else:
					firstc, lastc = item[0], item[-1]
					nextt = None if num + 1 >= count else tokens[num + 1]
					# skip_blank = lastout_ws or line_empty( output, '{(' ) or firstc in (WHITESPACE_CHAR + '({,;/')
					skip_blank = lastout_ws or firstc in (
								WHITESPACE_CHAR + ',;/')  # '({,;/')
					if nextt is None:
						append_blank = False
					elif not isinstance(nextt, str):  # int or float
						append_blank = True
					else:
						append_blank = lastc in ',;=' \
									   and nextt and nextt[0] in '(){}'
				output += '{}{}{}'.format('' if skip_blank else ' ',
										  item, ' ' if append_blank else '')
				lastout_ws = append_blank or output[-1] in WHITESPACE_CHAR
		return _gen_eff_data(output.replace(TAB, ' ' * TABSPACE))
	except Exception as exc:
		if DEBUG:
			print('_create_eff_data, Error: {}'.format(exc))
			traceback.print_exc()
			pdb.set_trace()
		else:
			print('{}\n_create_eff_data, name = {}, item = {}'.format(exc, name,
																	 item))

def loadCurrentData(fp, inOxz=False):
	"""load existing shipdata from telescope oxp/oxz"""

	global CurrentEffectData

	data = fp.read()
	if len(data) == 0:
		_log_error( 'loadCurrentData, failed to read file; length of 0: {}'
					.format(path) )
		return
	if inOxz: # zipfiles return bytes, not str
		data = data.decode( encoding = "utf-8", errors = "strict" )
	if len(data.strip()):
		CurrentEffectData = data.replace('\r', '').replace(TAB, ' ' * TABSPACE)
	else:
		_log_error( 'loadCurrentData, file failed to parse!' )

def loadEssentialData(fp, inOxz=False):
	"""load required shipdata from telescope oxp/oxz"""

	global Required_size, TelescopeEffectData

	data = fp.read()
	length = len(data)
	if length == 0:
		_log_error( 'loadEssentialData, failed to read file; length of 0: {}'
					.format(path) )
		return
	if inOxz: # zipfiles return bytes, not str
		data = data.decode( encoding = "utf-8", errors = "strict" )
	# don't decode, just output as is, after striping enclosing braces
	parsed = data[data.find('{') + 1 : data.rfind('}')]
	if len(parsed.strip()):
		TelescopeEffectData = parsed.replace('\r', '').replace(TAB, ' ' * TABSPACE)
		Required_size = len(TelescopeEffectData)
	else:
		_log_error( 'loadEssentialData, file failed to parse!' )

def getIdentifier(lines, inOxz=False):
	"""return the value of the 'identifier' key passed in lines,
	   which is the contents of a manifest.plist file"""

	for line in lines:
		if inOxz:  # zipfiles return bytes, not str
			line = line.decode(encoding="utf-8", errors="strict")
		match = IDENTIFIER_RE.match(line)
		if match:
			return match.group('id')

def loadFolder(path, isResouces=False, recurse=False):
	"""search 'path' for shipdata.plist files and load contents as effectdata
	   arg 'isResouces' is True when called for Oolite's Resouces folder
	   arg 'recurse' is True when called for the 'AddOns' folder, which can
	     have a tree of .oxp folders (while ManagedAddOns cannot)"""

	global TelescopeOxp

	def rptFolder():
		if LogFoldersSearched:
			if root not in reportedFolders:
				_log_message('\nfolder {}'.format(root))
				reportedFolders.append(root)
				print('  folder {}'.format(root))

	eRoot = eFile = dataFile = None
	err = False
	# Resouces -> /Config
	# AddOns -> /*.oxp/Config + any .oxz (/Config in oxz)
	# ManagedAddOns -> /*.oxp/Config + any .oxz (/Config in oxz)
	try:
		reportedFolders = []
		for root, dirs, files in os.walk(path, topdown=True):
			eRoot = root
			lowRoot = root.lower() if platformIsWindows else root
			inOxpFolder =  lowRoot.endswith(OXP_FOLDER_EXT)
			hasConfig = INPUT_FOLDER in dirs
			if (isResouces or inOxpFolder) and not hasConfig:
				if recurse:
					print('loadFolder, recursing as no Config folder in', dirs)
					continue
				else:
					print('loadFolder, bailing as no Config folder in', dirs)
					break
			dataFile = oxzs = None
			manifest = os.path.join(root, MANIFEST_FN)
			if os.path.exists(manifest):
				with open(manifest, 'r') as mp:
					lines = mp.readlines()
				if getIdentifier(lines) == TELESCOPE_ID:
					TelescopeOxp = root
					data = os.path.join(root, REQUIRED_FOLDER, REQUIRED_FN)
					with open(data, 'r') as fp:
						loadEssentialData(fp)
					data = os.path.join(root, OUTPUT_FOLDER, OUTPUT_FN)
					with open(data, 'r') as fp:
						loadCurrentData(fp)
			if isResouces or inOxpFolder:
				dataFile = os.path.join(root, INPUT_FOLDER, INPUT_FN)
				if not os.path.exists(dataFile):
					dataFile = None
			if not isResouces:
				oxzs = [fi for fi in files if fi.endswith(OXZ_FILE_EXT)]
				if len(oxzs) == 0:
					oxzs = None
			rptFolder()
			if dataFile:
				with open(dataFile) as fp:
					err = _load_effects(fp, dataFile) or err
			if oxzs:
				for file in oxzs:
					eFile = file
					err = _load_zipped_effects(root, file) or err

			if not recurse:
				break
			dirs[:] = [dr for dr in dirs if dr.endswith(OXP_FOLDER_EXT)]

	except IOError as msg:
		_log_warning('loadFolder, dataFile = "{}"'.format(dataFile))
		_log_warning('loadFolder, eRoot = "{}", eFile = "{}"'
					 .format(eRoot, eFile))
		_log_error('	{}'.format(msg) + NL)
		err = True
	except Exception as exc:
		if DEBUG:
			print('loadFolder, Error: {}'.format(exc))
			traceback.print_exc()
			pdb.set_trace()
		else:
			print('{}\nloadFolder, dataFile = "{}"'.format(exc, dataFile))
			print('	eRoot = "{}", eFile = "{}"'.format(eRoot, eFile))
		err = True
	return err

def ouput_update(output):
	try:
		fallback = None
		if TelescopeOxp:
			fname = os.path.join(TelescopeOxp, OUTPUT_FOLDER, OUTPUT_FILE)
			currFn = os.path.join(fname + OUTPUT_EXT)
			prevFn = os.path.join(fname + '.prev' + OUTPUT_EXT)
			fp = previousVersion(currFn, prevFn)
			fp.write(output)
			fp.close()
		else:
			if not zipfile.is_zipfile(TelescopeOxz):
				raise ValueError('update_eff_data, not a zip file: "{}"'
								 .format(TelescopeOxz))

			# make a temp folder to extract/create oxz
			tdStr = datetime.datetime.now().strftime("%H%M%S_%d%m%Y")
			tmpFolder = tempfile.mkdtemp(suffix='_' + tdStr, prefix='Ootmp_',
										 dir=Launch_dir)
			effects = '{}/{}/{}'.format(tmpFolder, OUTPUT_FOLDER, OUTPUT_FN)
			previous = '{}/{}/{}.prev{}'.format(tmpFolder, OUTPUT_FOLDER,
												OUTPUT_FILE, OUTPUT_EXT)
			shutil.unpack_archive(TelescopeOxz, tmpFolder, 'zip')

			# rename current version for possible fallback
			fallback = TelescopeOxz + '.cag'
			if Python2:
				os.rename(TelescopeOxz, fallback)
			else:
				os.replace(TelescopeOxz, fallback)

			# apply new data
			with previousVersion(effects, previous) as fp:
				fp.write(output)

			# create new oxz
			shutil.make_archive(TelescopeOxz, 'zip', root_dir=tmpFolder,
								base_dir=None)
			if Python2:
				os.rename(TelescopeOxz + '.zip', TelescopeOxz)
			else:
				os.replace(TelescopeOxz + '.zip', TelescopeOxz)
		return fallback
	except IOError as exc:
		msg = 'ouput_update, {}'.format(exc)
		msg += '  launched in {}, cwd = {}'.format(Launch_dir, os.getcwd())
		msg += '  TelescopeOxp = {}{}  TelescopeOxz = {}'.format(
				None if TelescopeOxp is None else TelescopeOxp, NL,
				None if TelescopeOxz is None else TelescopeOxz)
		print(msg)
		Logfile.write(msg + NL)
		if DEBUG:
			traceback.print_exc()
			pdb.set_trace()
	except Exception as exc:
		msg = 'ouput_update, {}'.format(exc)
		print(msg)
		Logfile.write(msg + NL)
		if DEBUG:
			traceback.print_exc()
			pdb.set_trace()

def reportResult():
	new_ents = len(Registered_ents)
	print(NL + 'Collected {} ents from {} files, output is {}'.format(
					new_ents, Input_file_count, Output_size))
	Logfile.write('\nProcessed {} files, new entities registered: {}'.format(
								Input_file_count, new_ents))
	msg = ''
	if Errors_in_log:
		msg += 'Encountered {} error{}.'.format(Errors_in_log,
							'' if Errors_in_log == 1 else 's') + NL
	if Warnings_in_log:
		msg += 'Issued {} warning{}.'.format(Warnings_in_log,
							'' if Warnings_in_log == 1 else 's') + NL
	print(msg)
	Logfile.write(NL + msg)

	Logfile.write( NL + '  input:{}shipdata {:>11,}'.format(
						' ' * 4, Input_size) + NL)
	generated = Output_size - Required_size
	Logfile.write(NL + ' output:{}required {:>11,}'.format(
						' ' * 4, Required_size) + NL)
	Logfile.write('{}generated {:>10,}'.format(' ' * 12, generated) + NL)
	Logfile.write('{}total {:>14,}'.format(' ' * 12, Output_size) + NL)

def update_eff_data():
	"""load all shipdata from oxp's to be loaded; output it as effectdata"""
	err = False
	update = True
	fallback = tmpFolder = None
	output = ''
	try:
		output += _gen_eff_data(OUTPUT_FILE_HEADER)

		err = loadFolder(OoFolders['appDir'], isResouces=True)
		err = loadFolder(OoFolders['maoDir']) or err
		err = loadFolder(OoFolders['aoDir'], recurse=True) or err

		output += _create_eff_data()
		output += _gen_eff_data(TelescopeEffectData)
		output += _gen_eff_data(OUTPUT_FILE_TAIL)

		if TelescopeOxp is None and TelescopeOxz is None:
			msg = '** Failed to locate Telescope oxp, cannot update.'
			print(msg)
			Logfile.write(NL + msg + NL)
			update = False
			err = True

		reportResult()
		# compare to existing to see if effectdata has changed
		if output.strip() == CurrentEffectData.strip():
			msg = 'No change in effect data, halting Telescope update.'
			print(msg)
			Logfile.write(NL + msg + NL)
			update = False

		# output results to installed Telescope
		if update:
			fallback = ouput_update(output)
			msg = 'Be sure to hold down the Shift key the next time you start Oolite, '
			msg += '\nso the new effects are loaded.' + NL
			print(msg)
			Logfile.write(NL + msg + NL)

	except Exception as exc:
		print('update_eff_data, {}'.format(exc))
		if DEBUG:
			traceback.print_exc()
			pdb.set_trace()
	else:
		# clean up - won't execute if any exception was raised
		if fallback and os.path.exists(fallback):
			os.remove(fallback)
	finally:
		if tmpFolder and os.path.exists(tmpFolder):
			shutil.rmtree(tmpFolder)
		if update and err:
			print( 'check log file for errors' )
	return err, update

OoFolders = {'appDir': None, 'maoDir': None, 'aoDir': None}
def findOolite(cwd):
	"""locate closest oolite.app, AddOns & ManagedAddOns folders"""

	try:
		def foundAll():
			return OoFolders['appDir'] and OoFolders['maoDir'] and OoFolders['aoDir']

		def checkFolder(path, curr):
			if curr == 'Resources' and OoFolders['appDir'] is None \
					and path.endswith('oolite.app'):
				OoFolders['appDir'] = os.path.join(path, curr)
			if curr == 'ManagedAddOns' and OoFolders['maoDir'] is None \
					and path.endswith('Oolite'):
				OoFolders['maoDir'] = os.path.join(path, curr)
			if curr == 'AddOns' and OoFolders['aoDir'] is None:
				# no path test as we can find AddOns folder in different
				# places (Linux) and location(s) varies between OS's
				OoFolders['aoDir'] = os.path.join(path, curr)

		def checkChildren(path):
			for child, dirs, _ in os.walk(path):
				if OoFolders['appDir'] and OoFolders['appDir'] == child:
					dirs[:] = []	# found appDir, cancel this branch
				if OoFolders['maoDir'] and OoFolders['maoDir'] == child:
					dirs[:] = []	# found maoDir, cancel this branch
				if OoFolders['aoDir'] and OoFolders['aoDir'] == child:
					dirs[:] = []	# found aoDir, cancel this branch
				if platformIsWindows and child.endswith('AppData'):
					dirs[:] = []	# skip AppData (large & n/a)
				for fldr in dirs:
					checkFolder(child, fldr)
					if foundAll():
						return True
			return False

		# check child folders
		if checkChildren(cwd):
			return True

		# check parent folders
		drive, _ = os.path.splitdrive(cwd)
		driveRoot = os.path.join(drive, os.sep)
		# for .split(path): "if path ends in a slash, tail will be empty"
		head = cwd[:-len(os.sep)] if cwd.endswith(os.sep) else cwd
		while True:
			head, _ = os.path.split(head)
			# exclude root so not searching entire drive
			if head == driveRoot or head == os.sep or head == '':
				break
			if checkChildren(head):
				return True

		# check user folder
		userDir = os.path.expanduser('~' + getpass.getuser())
		if checkChildren(userDir):
			return True

		# check elsewhere
		if platformIsWindows:
			# check for root installation
			rootDirs = []
			for rd in os.listdir(driveRoot):
				dirPath = os.path.join(drive, os.sep, rd)
				if not os.path.isdir(dirPath):
					# skip shortcuts to folders (are returned by listdir!)
					continue
				lowRD = rd.lower()
				if lowRD in ['$recycle.bin', 'boot', 'config.msi',
							 '$winreagent', 'system volume information']:
					continue
				if lowRD.startswith('$windows.'):
					continue
				try:
					if 'oolite.app' in os.listdir(dirPath):
						rootDirs.append(dirPath)
				except PermissionError:
					pass
				except Exception as exc:
					if DEBUG:
						print('findOolite, Error: {} for "{!r}"'.format(exc, dirPath))
						traceback.print_exc()
						pdb.set_trace()
			for root in rootDirs:
				if checkChildren(os.path.join(driveRoot, root)):
					return True
			# check for installation as an app
			progFiles = os.path.join(drive, os.sep, 'Program Files')
			if os.path.exists(progFiles):
				for check in os.listdir(progFiles):
					if checkChildren(os.path.join(progFiles, check)):
						return True
			progFiles = os.path.join(drive, os.sep, 'Program Files (x86)')
			if os.path.exists(progFiles):
				for check in os.listdir(progFiles):
					if checkChildren(os.path.join(progFiles, check)):
						return True
		elif platformIsLinux:
			appDir = os.path.join(os.sep, 'usr', 'local')
			if checkChildren(appDir):
				return True
			appDir = os.path.join(os.sep, 'opt')
			if checkChildren(appDir):
				return True
			appDir = os.path.join(os.sep, 'home', 'user')
			if checkChildren(appDir):
				return True
			appDir = os.path.join(os.sep, 'usr')
			if checkChildren(appDir):
				return True
		elif platformIsMacOS:
			appDir = os.path.join(os.sep, 'Applications')
			if checkChildren(appDir):
				return True
			appDir = os.path.join(os.sep, 'Library')
			if checkChildren(appDir):
				return True
		else: # some other OS that's not supported
			print('findOolite, unsupported platform "{}", aborting'.format(sys.platform))
			sys.exit(1)

		return False
	except Exception as exc:
		if DEBUG:
			print('findOolite, Error: {}'.format(exc))
			traceback.print_exc()
			pdb.set_trace()

def previousVersion(currFn, prevFn):
	if os.path.exists(currFn):
		if os.path.exists(prevFn):
			os.remove(prevFn)
		if Python2:
			os.rename(currFn, prevFn)
		else:
			os.replace(currFn, prevFn)
	if Python2:
		fp = open(currFn, mode='w')
	else:
		fp = open(currFn, mode='w', encoding='utf-8')
	return fp

def openLogFile():
	"""open both output and log files, preserving any existing as .prev"""

	global Logfile
	opened = False
	try:
		Logfile = previousVersion(os.path.join(Launch_dir, LOG_FN),
								  os.path.join(Launch_dir,
											   LOG_FILE + '.prev' + LOG_EXT))
		opened = True
	except Exception as exc:
		print('openLogFile, for "{}": {}  {}'.format(LOG_FN, NL, exc) )
	return opened

def main():
	global Launch_dir

	err = True
	Launch_dir = os.getcwd()
	print('Launched in:', Launch_dir)
	try:
		found = findOolite(Launch_dir)
		print('main, {} all 3 Oolite folders'.format('found' if found else 'failed to find'))
		print(NL.join('{}{}: {}'.format(' ' * (8 - len(k)), k, v) for k, v in OoFolders.items()))
		print('')
		if not found or not openLogFile():
			if DEBUG:
				pdb.set_trace()
			sys.exit(1)

		err, update = update_eff_data()
	except Exception as exc:
		if DEBUG:
			print('main, Error: {}'.format(exc))
			traceback.print_exc()
			pdb.set_trace()
	finally:
		Logfile.close()

	# noinspection PyBroadException
	try:
		msg = 'Details are in log file.' + NL
		msg += 'Hit any key to close window.'
		input(msg)
	except:
		pass
	return err

if __name__ == "__main__":
	sys.exit(1 if main() else 0)
